begin EQU &8000
beginhibyte EQU begin/&100
opcodelobyte EQU &D6

ORG begin

;********************************************************
;*	A MINIMAL (0,5K) TEXT BASED SNAKE VERSION.	*
;*	 PROGRAMMING EXERCISE FOR THE AMSTRAD CPC	*
;*			USING FIRMWARE ROUTINES		*
;********************************************************

;inspired by youtu.be/ExwqNreocpg , I tried to program a Snake
;clone as small as possible. I ended up at exactly 512 bytes.

;THE GAME:
;Navigate the snake with n (left) and m (right) and exit with ESC.
;If you collide with the border, eat poisonous skulls or byte your
;own body, you lose and a new game will start.
;Eat food to gain points and get longer (by the indicated digits).
;But beware, every time you eat, the game also becomes faster. (The
;acceleration is always the same, no matter how many points you ate).
;During a session, the program also keeps track of the highscore, but
;it doesn't save it, unfortunately.

;PHILOSOPHY:
;The main goal here was small size, not speed! But I wanted to have:
;*	A decent "sprite" set for the snake (different tiles for straight and
; 	turned body, a living and a dead head and all of that in 4 directions.)
;*	Score and Highscore (not saved, unfortunately)
;*	Nice custom colors
;*	Decent random numbers
;*	A nice game title made of snakes. (I had to pass on this one...)

;Except for the random seed and the table with the coordinates
;of the snake body parts, all game variables are stored in
;registers, so a lot of PUSH and POP has to happen, which takes
;less bytes of opcodes to program than accessing memory directly:
; 	BC = length of snake
; 	D = direction of movement (0/1/2/3 = clockwise up/right/down/left)
; 	E = speed (delay = E/2 framerates)
; 	HL = current coordinates of snake head
; 	IXH = High Score

;Also, I used plenty of Firmware routines instead of things
;like own sprite routines. So, this version is not very fast, but
;the main goal was the small size.

;I tried to use as many optimization tricks as possible, including:
;*	Self modifying code.
;*	Rotating the snake sprites in the code instead of hardcoding them.
;*	Reusing register values accidentally left over by firmware.
;*	Other optimization tricks (JP instead of CALL&RET, etc.)

;SYSTEM CALLS
	scrwaitflyback EQU &BD19
	scrsetmode EQU &BC0E		;A=mode, (AFBCDEHL corrupt)
	scrfillbox EQU &BC44		;HL=top left, DE=bottom right (text coords)
	scrsetink EQU &BC32 		;A=ink, BC=color
	scrsetborder EQU &BC38		;BC=color
	kbdreadkey EQU &BB09 		;A=char, C=true if key pressed
	txtwritechar EQU &BB5A 		;A=char, all registers preserved
	txtsetcursor EQU &BB75 		;H=x, L=y, AF & HL are corrupt
	txtsetcolumn EQU &BB6F		;A=x, HL corrupt
	txtsetpen EQU &BB90 		;A=pen, AF & HL corrupt
	txtreadchar EQU &BB60 		;char in A if C is true
	txtgetmatrixadress EQU &BBA5	;A=symbol, adr returns in HL

;CONSTANTS
;graphics
	width EQU 36: height EQU 16	;dimensions of game table
	borderoffset EQU width*&100+height
	topleft EQU &0203
	bottomright EQU topleft+borderoffset
;key codes
	mvleftkey EQU "n"
	mvrightkey EQU "m"
	escapekey EQU &FC
;symbol codes
	blank EQU 32
	snakeheads EQU 240 	;+ direction = right orientation
	snakebodies EQU 244 	; = " =
	snaketurns EQU 248 	; = " =
	deadhead EQU 252
	food EQU "1" 		;+ offset (points-1)
	skull EQU deadhead+2
;game constants
	maxpoints EQU 4		;has to be a square of 2
	numberofitems EQU 8 	;how many food & skulls at the beginning
	initialdelay EQU 50 	;x/2 framerates delay

;INITIALISATION

	;LOAD CUSTOM SYMBOLS
	;The symbols are hardcoded in just one orientation.
	;Rotating them by the program itself saves us about 50 bytes.
	;(This is horribly long but I saved 4 more bytes by not using a subroutine.)
	LD A,snakeheads				;custom font begins at this symbol
	CALL txtgetmatrixadress			;here's where we have to write the new symbols
	PUSH HL					;save it for the real routine
	LD B,8*16				;we have 16 custom symbols
	eraseloop:				;first, we clean up (0s all over the place)
		LD (HL),A			;(A=0) because the rotation routine works with OR
		INC HL
		DJNZ eraseloop
	POP HL					;reload the destination memory
	LD DE,symbols				;pointer to the (hardcoded) base symbols
	foursymbols: 				;...they have to be rotated AND copied
		PUSH DE				;(save current symbol for end of loop)
 		LD IXH,4			;we rotate each symbol 4 times clockwise
		rotation:			;DE=source, HL=destination!
			PUSH HL			;save current destination adress for later
			INC B			;mask for the destination bit (begin with 1 = bit 0)
  			rbytelp:		;loop through the lines (bytes) of source symbol
  				POP HL:PUSH HL
				LD C,&80		;mask for the checked bit (we begin with bit 7)
				rbitlp:			;loop through the bits of current line
					LD A,(DE)	;byte to copy/rotate
					AND C		;check if this bit is set
					JR Z,nextbit
					LD A,B		;set the corresponding bit
					OR (HL)		;in the corresponding byte
					LD (HL),A	;and load it back
					nextbit:
					INC HL		;ext dest byte
					SRL C		;shift to next bit for source
					JR NZ,rbitlp
				INC DE			;next source byte
				SLA B			;shift to next bit for dest
				JR NZ,rbytelp
			POP DE 				;change source adress to symbol we just rotated
			DEC IXH
			JR NZ,rotation		;next 3 rotations.
		POP DE					;After having rotated a symbol 4 times, we go
		LD A,8:	ADD E: LD E,A	;"back" to the next hardcoded one.
		JR NZ,foursymbols		;symbols end at &200, so loop until DE=0

	;SET COLORS
	;Here, I use a lot of values left over by the firmware routines
	;in order to save some bytes. Admittedly, the color scheme is also
	;a bit influenced by the selection of values I could choose from...
	PUSH BC						;(A and BC are still 0 from rotation routine)
	CALL scrsetink 				;ink 0,0
	POP BC
	call scrsetborder 			;border 0
	LD A,2:	INC B 				;B & C happen to be 20 here, so:
	call scrsetink 				;ink 2,19,20 (poisonous flash)
	LD A,C: LD B,L: LD C,L			;C % 4 = 3, L % &20 = 8
	call scrsetink				;ink 3,8 (inviting magenta)
	LD A,L: call scrsetmode			;L % 4 = 1
	INC A: LD B,H: LD C,H			;A was 0, E % &20 = 24
	call scrsetink				;ink 1,24 (snakey yellow)

;MAIN PROGRAM LOOP
programloop:
	;IXH, the high score is already 0 after the rotation routine
	CALL maingame
	CP escapekey-mvleftkey 			;exit if ESC pressed (but we have an
	RET Z					;offset from the evaluation before)
	CALL setcursor_readchar			;in all other cases:
	LD A,deadhead: ADD D
	CALL txtwritechar			;GAME OVER
	CALL delayloop 				;A=&FC after writechar. So we enter right
						;into the loop of the delay sub
	JR programloop				;next game

;MAIN GAME LOOP
maingame:

	;DRAW BORDER / CLEAR FIELD
	LD A,245				;striped in ink 1 & 3
	LD HL,topleft-&101
	LD DE,bottomright
	CALL scrfillbox
	XOR A					;ink 0 (erase)
	LD HL,topleft
	LD DE,bottomright-&101
	CALL scrfillbox

	;PUT POINTS AND SKULLS ON SCREEN
	LD B,numberofitems			;counter
	itemloop:
		CALL skullandfood		;lay out skull & points
		DJNZ itemloop
	;SET GAME VARIABLES
	;(register A happens to be 2 at this point, that's convenient)
	LD C,A					;BC = length*2 (B is already 0 here)
	PUSH BC					;unfortunately we have to backup
	CALL printscore				;for score output
	INC A: CALL setpen_getrandomXY		;set pen 1 & create random starting point
	LD A,R:	AND 3: LD D,A			;random moving direction
	LD E,initialdelay

	;LOOP WHILE SNAKE IS ALIVE
	movingloop:
		POP BC					;restore snake length
		CALL setcursor_readchar			;check if something is in the way
		RET NC					;unknown symbol = border -> Game Over
		CP snakeheads: RET NC			;own body or skull -> Game Over
		CP "0": CALL NC,foodfound 		; > "0" -> snake found food
							;(else crawl on)
		drawsnakehead:
		LD A,snakeheads				;D is used as offset to get the body
		ADD D 					;parts in the right orientation
		CALL txtwritechar

		;WRITE COORDS OF SNAKE HEAD TO TOP OF TABLE
		LD IY,SnakeCoordsTable
		ADD IY,BC				;(base+snake length)
		LD (IY),H: LD (IY+1),L
		CALL setcursor_readchar 		;position cursor back to head position
		
		;WAIT AND READ KEYS
		CALL delayandinkey

		;EVALUATE KEYS
		JR NC,movingforward 			;if no key pressed, keep moving
		SUB mvrightkey: JR Z,doturnright
		DEC A: JR Z,doturnleft			;(left arrow key code = right arrow+1)
		CP escapekey-mvleftkey: RET Z 		;exit if ESC pressed
							;Any other key = not valid, keep moving:
		movingforward:
		LD A,snakebodies-snaketurns		;use straight body. For optimization
		ADD D					;we have to subtract turned body index:
		drawbody:
		ADD snaketurns				;placed here instead of at the end of
							;the doturn-subs for less code
		CALL txtwritechar 			;draw the body (over the head. New head
							;will be drawn in the next round
							;after position calculation)
		PUSH BC: PUSH HL: PUSH DE  		;backup game variables
		;SHIFT SNAKE COORDS IN MEMORY
		LD HL,SnakeCoordsTable+2
		LD DE,SnakeCoordsTable 			;(BC is len*2 already)
		PUSH DE 				;for POPping it into HL afterwards
		LDIR
		;ERASE TAIL (ON SCREEN)
		POP HL 					;(=SnakeCoordsTable)
		LD D,(HL): INC HL 			;load tail coords
		LD E,(HL): EX HL,DE			;(first pair in table)
		CALL txtsetcursor
		LD A,blank: CALL txtwritechar 		;overwrite tail
		POP DE 					;restore direction (& delay)
		;CALCULATE NEW COORDINATES (from D=direction)
							;using self modifying code
		LD A,D:	ADD opcodelobyte 		;using variables for easier
		LD L,A: LD H,beginhibyte+1		;relocating the code
		LD A,(HL)				;read the opcode
		LD (turninstruction),A			;load it to the program
		POP HL					;restore old coordinates
		turninstruction:
			NOP				;and calculate new ones
			JR movingloop

doturnright:
	LD A,D:	INC A: AND 3: LD D,A			;next value clockwise, 4->0
	JR drawbody					;offset for turns will be added there

doturnleft:
	LD A,D: DEC A: AND 3: LD D,A			;next value counterclockwise, -1 -> 3
	DEC A: AND 3					;A-1 gives us symbol for left turn
	JR drawbody					;offset for turns will be added there

foodfound:
	SUB "0"						;calculate points, A = counter
	PUSH HL: PUSH DE 				;save game vars
	stretchsnakeloop:				;loop for every point added
		INC BC: INC BC				;increase length (coords are in pairs)
		PUSH BC
		LD HL,SnakeCoordsTable
		ADD HL,BC				;calc old top adress of table
		PUSH HL: POP DE				;copy to DE
		INC DE: INC DE				;calc new top adress of table
		LDDR					;shift coords up by one pair
		POP BC					;restore BC
		DEC A
		JR NZ,stretchsnakeloop
	CALL putfood
	PUSH BC: CALL printscore			;display new score
	POP BC: POP DE: POP HL				;restore vars
	DEC E 						;decrease delay (go faster)
	JR setcursor_readchar 				;(because we'll draw the snakehead after
							;returning automatically)

skullandfood:
	LD A,2:	CALL setpen_getrandomXY			;skulls are pen 2, random coords
	LD A,skull: CALL txtwritechar 			;write skull & continue to
putfood: 						;put new food somewhere
	LD A,3: CALL setpen_getrandomXY			;food is pen 3
	LD A,R: AND maxpoints-1 			;reduce random to 0-(maxpoints-1)
	ADD "1"						;add char offset for correct symbol
	CALL txtwritechar
	LD A,1:JP txtsetpen 				;back to pen 1 (A=&31) for snake
							;(JR instead of CALL / RET)
setpen_getrandomXY:
	CALL txtsetpen					;set pen from reg A
	getx: 	call getrandom
			CP width-2
			JR NC,getx			;only accept X in our range
	ADD 4 						;add offset for centered gametable
	PUSH AF						;buffer to stack because H will be ruined
	gety: 	call getrandom
			CP height-2
			JR NC,gety			;only accept Y in our range
	POP HL 						;restore H from buffer
	ADD 5: LD L,A					;and load L (with offset)
	CALL setcursor_readchar
	CP blank 					;check if space free
	JR NZ,getx					;if not try other coords
	RET

getrandom:
	;Routine from CPCWiki.eu - Thank you, guys!
	LD HL,(randseed)
	LD A,R: LD D,A: LD E,A: ADD HL,DE
	XOR L: ADD A: XOR H: LD L,A
	LD (randseed),HL
	RET

setcursor_readchar:
	PUSH HL				;conserve HL
	call txtsetcursor
	POP HL
	JP txtreadchar			;optimization - sometimes we need to read the char
					;and if not, it doesn't hurt (RET automatically)

delayandinkey:
	LD A,E: SRL A			;we only wait half the amount
	delayloop:
		CALL scrwaitflyback
		DEC A
		JR NZ,delayloop
	JP kbdreadkey			;after delay, we read keyboard buffer
					;(and return automatically)

printscore:
	;EVALUATE AND PRINT POINTS
	;actually only works up to 255 points, but more are not likely
	LD H,A: LD L,A: 			;locate 2,2 (A happens to be 2)
	LD IY,textlabels: call printword	;"Sc" - IY ends at begin of next word
	LD A,B: RRA: LD A,C: RRA		;length = BC/2
	CP IXH: JR C,printbothvalues		;check if Score>Hiscore
	LD IXH,A				;if so, save new Hiscore
	printbothvalues:
	CALL printbyteasdecimal			;print score
	LD H,34: INC L: call printword		;"Hi" (L happens to be 1)
	LD A,IXH				;load Hiscore and proceed automatically:
	printbyteasdecimal:
	;Routine from CPCWiki.eu - Thank you, guys!
	LD B,100: CALL printdecimaldigit
	LD B,10: CALL printdecimaldigit
	LD B,1
	printdecimaldigit:
	LD C,0
	decimaldivide:
	SUB B: JR C,displaydecimaldigit
	INC C: JR decimaldivide
	displaydecimaldigit:
	ADD A,B: PUSH AF: LD A,C: ADD A,"0"
	CALL txtwritechar: POP AF
	RET

printword: ;HL=coords, IY=adr, word ends with colon
	CALL txtsetcursor
	charloop:
		LD A,(IY)
		call txtwritechar
		INC IY
		CP ":"
		RET Z
		JR charloop

ORG begin+&100+opcodelobyte
turnopcodes:					;Table for SMC - four movements manipulating
	DEC L: INC H: INC L: DEC H		;the x and y coordinate according to D

textlabels:
	DB "Sc:", "Hi:"

symbols:
	;headalive
	DB &0C, &1E, &73, &FF, &FF, &73, &1E, &0C ;left
	;body straight
	DB &00, &FE, &BB, &BB, &BB, &BB, &FE, &00 ;left
	;body turning
	DB &7E, &FE, &E6, &B2, &BE, &9C, &F8, &00 ;left
	;headdead
	DB &0C, &12, &73, &9F, &9F, &73, &12, &0C

;gametitle - I'd love to use that, but I'm afraid there's no space
	;DB 253,249,254,253,254,244,253,249,254,250,253,251,253,251,251
	;DB 252,249,254,248,248,248,248,249,248,248,249,254,248,245,032
	;DB 252,249,255,248,252,255,248,032,248,248,032,250,252,251,251

randseed:
	DB 0,0

SnakeCoordsTable:
	DS initialdelay*2*maxpoints
